#==========================================================================
#
# VHD OFFLINE DEFRAGMENTATION
#
# AUTHOR: Dennis Span (http://dennisspan.com)
# DATE  : 11.10.2017
#
# COMMENT 1: This script mounts a VHD(X) file, defrags it and unmounts it again
# COMMENT 2: Please add your custom parameters on the command line when calling the script. Only the parameter '-File', containing the path and file name of the VHD(X) file, is mandatory
#            
#            Examples:
#            powershell.exe -ExecutionPolicy ByPass -File "C:\Temp\VHD Offline Defrag.ps1" -VHDFile "C:\Temp\MyVirtualHardDisk.vhdx"
#            powershell.exe -ExecutionPolicy ByPass -File "C:\Temp\VHD Offline Defrag.ps1" -VHDFile "C:\Temp\MyVirtualHardDisk.vhdx" -Logdir "C:\MyLogDir" -Packagename "DefragMyVHDFile"
#            powershell.exe -ExecutionPolicy ByPass -File "C:\Temp\VHD Offline Defrag.ps1" -VHDFile "C:\Temp\MyVirtualHardDisk.vhdx" -DefragFile "C:\Temp\Defraggler.exe" -Logdir "C:\MyLogDir" -Packagename "DefragMyVHDFile"
#        
# Change log:
# -----------
# 16.10.2017 Dennis Span: added -> check if VHD(X) file exists
# 17.11.2017 Dennis Span: added -> set the Unique Disk ID back to the original one before unmounting the VHD(X) file
#                         Note: when mounting the VHD(X) file on the same machine where it was created, a disk collision occurs as soon as the VHD(X) file
#                         is taken online. As soon as the disk is online, a new Unique Disk ID is set, which makes the VHDX file unbootable.
#==========================================================================

param (
    [Parameter(Mandatory=$true)]
    [string]$VHDFile,                                        # [edit] enter the path and name of the VHD or VHDX file that needs to be defragged (e.g. "c:\Temp\MyVirtualHardDisk.vhdx", "\\MyFileServer\MyFileShare\MyVirtualHardDisk.vhdx")
    [Parameter(Mandatory=$false)]
    [string]$DefragFile = "defrag.exe",                      # [edit] enter the path (= optional) and executable name of the defragger program (e.g. 'defrag.exe', 'C:\Temp\df.exe' (= Defraggler), etc.)
    [Parameter(Mandatory=$false)]
    [string]$LogDir = "C:\Logs",                             # [edit] enter the path to your log directory (e.g. "C:\Logs")
    [Parameter(Mandatory=$false)]
    [string]$PackageName = "VHDOfflineDefrag"                # [edit] enter the package name (the name of the log directory and log file is based on the package name)
)

# Custom variables retrieved from the command line parameters
$VHDFileToDefrag = $VHDFile
$DefragFile = $DefragFile
$BaseLogDir = $LogDir
$PackageName = $PackageName

# define Error handling
# note: do not change these values
$global:ErrorActionPreference = "Stop"
if($verbose){ $global:VerbosePreference = "Continue" }

# FUNCTION DS_WriteLog
#==========================================================================
Function DS_WriteLog {
    <#
        .SYNOPSIS
        Write text to this script's log file
        .DESCRIPTION
        Write text to this script's log file
        .PARAMETER InformationType
        This parameter contains the information type prefix. Possible prefixes and information types are:
            I = Information
            S = Success
            W = Warning
            E = Error
            - = No status
        .PARAMETER Text
        This parameter contains the text (the line) you want to write to the log file. If text in the parameter is omitted, an empty line is written.
        .PARAMETER LogFile
        This parameter contains the full path, the file name and file extension to the log file (e.g. C:\Logs\MyApps\MylogFile.log)
        .EXAMPLE
        DS_WriteLog -InformationType "I" -Text "Copy files to C:\Temp" -LogFile "C:\Logs\MylogFile.log"
        Writes a line containing information to the log file
        .Example
        DS_WriteLog -InformationType "E" -Text "An error occurred trying to copy files to C:\Temp (error: $($Error[0]))" -LogFile "C:\Logs\MylogFile.log"
        Writes a line containing error information to the log file
        .Example
        DS_WriteLog -InformationType "-" -Text "" -LogFile "C:\Logs\MylogFile.log"
        Writes an empty line to the log file
    #>
    [CmdletBinding()]
	Param( 
        [Parameter(Mandatory=$true, Position = 0)][ValidateSet("I","S","W","E","-",IgnoreCase = $True)][String]$InformationType,
        [Parameter(Mandatory=$true, Position = 1)][AllowEmptyString()][String]$Text,
        [Parameter(Mandatory=$true, Position = 2)][AllowEmptyString()][String]$LogFile
    )

	$DateTime = (Get-Date -format dd-MM-yyyy) + " " + (Get-Date -format HH:mm:ss)
	
    if ( $Text -eq "" ) {
        Add-Content $LogFile -value ("") # Write an empty line
    } Else {
	    Add-Content $LogFile -value ($DateTime + " " + $InformationType.ToUpper() + " - " + $Text)
    }
}
#==========================================================================

################
# Main section #
################

# Disable File Security
$env:SEE_MASK_NOZONECHECKS = 1

# Global variables
$StartDir = $PSScriptRoot # the directory path of the script currently being executed
$LogDir = (Join-Path $BaseLogDir $PackageName).Replace(" ","_")
$LogFileName = "Config_$($PackageName).log"
$LogFile = Join-path $LogDir $LogFileName

# Create the log directory if it does not exist
if (!(Test-Path $LogDir)) { New-Item -Path $LogDir -ItemType directory | Out-Null }

# Create new log file (overwrite existing one)
New-Item $LogFile -ItemType "file" -force | Out-Null

DS_WriteLog "I" "START SCRIPT - Config $PackageName" $LogFile
DS_WriteLog "-" "" $LogFile


#####################################
# Defrag VHD(X) offline             #
#####################################

DS_WriteLog "I" "==========================" $LogFile
DS_WriteLog "I" "Defrag VHD(X) (offline)   " $LogFile
DS_WriteLog "I" "==========================" $LogFile

DS_WriteLog "-" "" $LogFile
DS_WriteLog "I" "File to defrag: $VHDFileToDefrag" $LogFile
DS_WriteLog "I" "Defrag executable: $DefragFile" $LogFile
DS_WriteLog "I" "Log file: $LogFile" $LogFile
DS_WriteLog "-" "" $LogFile

# Write verbose output
Write-Host "Defrag VHD(X) (offline)"
Write-Host "======================="
Write-Host ""
Write-Host "File to defrag: $VHDFileToDefrag"
Write-Host "Defrag executable: $DefragFile"
Write-Host "Log file: $LogFile"
Write-Host ""

# Check if the VHD(X) file exists
DS_WriteLog "I" "Check if the VHD(X) file '$VHDFileToDefrag' exists" $LogFile
if ( (Test-Path $VHDFileToDefrag ) -eq $True ) {
    DS_WriteLog "I" "The VHD(X) file '$VHDFileToDefrag' exists" $LogFile
} else {
    DS_WriteLog "E" "The VHD(X) file '$VHDFileToDefrag' does NOT exist or cannot be reached" $LogFile
    Write-Host "The script ended in an error. Please see the log file '$LogFile'"
    Exit 1
}
DS_WriteLog "-" "" $LogFile

# Retrieve drives before mounting the VHD(X)
$DrivesAvailableBeforeVHDMount = (Get-PSDrive -PsProvider FileSystem).Name
DS_WriteLog "I" "Retrieve available drives (before mount): $([string]$DrivesAvailableBeforeVHDMount)" $LogFile
DS_WriteLog "-" "" $LogFile

# Retrieve available disks before VHD(X) mount
$DisksAvailableBeforeVHDXMount = (Get-WmiObject Win32_DiskDrive).Index | sort-object
DS_WriteLog "I" "Retrieve available disks (before mount): $([string]$DisksAvailableBeforeVHDXMount)" $LogFile
DS_WriteLog "-" "" $LogFile

# Retrieve the disk number of the disk with the active partition
$DiskNumber = ((Get-WmiObject -Class Win32_DiskPartition | Where-Object { $_.Bootable -eq $True }).DeviceID).SubString(6,1)
DS_WriteLog "I" "Retrieve the disk number of the disk with the active partition: $($DiskNumber)" $LogFile
# Retrieve the unique disk ID (= signature) of the disk with the active partition
$UniqueDiskIDDec = (Get-WmiObject Win32_DiskDrive | Where-Object { $_.Index -eq $DiskNumber }).Signature
DS_WriteLog "I" "Retrieve the unique disk ID (= signature) of the disk with the active partition (decimal value): $($UniqueDiskIDDec)" $LogFile
# Convert the unique ID (= signature), which is returned in decimal notation, to hexidecimal notation
$UniqueDiskIDHex = '{0:X}' -f $UniqueDiskIDDec
DS_WriteLog "I" "Convert decimal value of the unique disk ID to hexidecimal: $($UniqueDiskIDHex)" $LogFile

DS_WriteLog "-" "" $LogFile

# Create temporary diskpart configuration file (to be able to mount the VHD(X))
$DiskpartFile = "$env:TEMP\diskpart_mount.txt"
$Line1 = "SELECT VDISK FILE=""$($VHDFileToDefrag)"""
$Line2 = "ATTACH VDISK"
$Line3 = "ONLINE DISK"
DS_WriteLog "I" "Create temporary diskpart file to mount the VHD(X) file ($DiskpartFile)" $LogFile
try {
    New-Item $DiskpartFile -type file -force  | Out-Null
    DS_WriteLog "I" "Temporary diskpart file was created successfully" $LogFile
    Add-Content $DiskpartFile $Line1
    DS_WriteLog "I" "The line $Line1 was added successfully" $LogFile
    Add-Content $DiskpartFile $Line2
    DS_WriteLog "I" "The line $Line2 was added successfully" $LogFile
    Add-Content $DiskpartFile $Line3
    DS_WriteLog "I" "The line $Line3 was added successfully" $LogFile
} Catch {
    DS_WriteLog "E" "An error occurred trying to create the diskpart configuration file (error: $($Error[0]))" $LogFile
    Write-Host "The script ended in an error. Please see the log file '$LogFile'"
    Exit 1
}

DS_WriteLog "-" "" $LogFile

# Mount VHD(X) (using diskpart)
DS_WriteLog "I" "Mount (attach) the VHD(X) (using diskpart.exe)" $LogFile
$Process = Start-Process -WindowStyle hidden -FilePath 'diskpart.exe' -ArgumentList "/S $DiskpartFile" -Wait -PassThru
$ProcessExitCode = $Process.ExitCode
switch ($ProcessExitCode)
{        
    0 { 
        DS_WriteLog "I" "The diskpart command executed successfully" $LogFile
    }
    -2147024809 { 
        DS_WriteLog "E" "An error occurred. Either the virtual disk is already attached or the file or directory is corrupted!" $LogFile
        DS_WriteLog "E" "Please detach the VHD(X) file and run this script again or try to attach the VHD(X) file manually in the Computer Management console" $LogFile
        DS_WriteLog "E" "Please make sure you are using a valid VHD(X) file containing a Windows operating system (Windows 7/Windows Server 2008 R2 or higher)" $LogFile
        Write-Host "The script ended in an error. Please see the log file '$LogFile'"
        Exit 1
    }
    default {
        DS_WriteLog "E" "An error occurred executing the diskpart command (error: $ProcessExitCode)" $LogFile
        Write-Host "The script ended in an error. Please see the log file '$LogFile'"
        Exit 1
    }
}

Write-Host "The VHD(X) file was mounted successfully"
DS_WriteLog "-" "" $LogFile

# Retrieve drives after mounting the VHD(X)
$DrivesAvailableAfterVHDMount = (Get-PSDrive -PsProvider FileSystem).Name
DS_WriteLog "I" "Retrieve available drives (after mount): $([string]$DrivesAvailableAfterVHDMount)" $LogFile
DS_WriteLog "-" "" $LogFile

# Check which drive letter or driver letters were added after mounting the VHD(X) file (drive letters are written without a colon; e.g. "D" instead of "D:")
try { 
    [array]$Drives = ((Compare-Object $DrivesAvailableBeforeVHDMount $DrivesAvailableAfterVHDMount).InputObject).ToUpper()
} catch {
    DS_WriteLog "E" "No additional drives were detected. It is possible that the VHD(X) file was mounted, but that no drive letter could be assigned" $LogFile
    DS_WriteLog "E" "Please make sure you are using a valid VHD(X) file containing a Windows operating system (Windows 7/Windows Server 2008 R2 or higher)" $LogFile
    Write-Host "The script ended in an error. Please see the log file '$LogFile'"
    Exit 1
}

# Continue check which drive letter or driver letters were added after mounting the VHD(X) file (drive letters are written without a colon; e.g. "D" instead of "D:")
$DriveCount = $Drives.Count
switch ($DriveCount)
{        
    0 { 
        DS_WriteLog "E" "No new drives were added. Apparently the VHD(X) mount did not succeed" $LogFile
        DS_WriteLog "E" "This script will now quit." $LogFile
        Write-Host "The script ended in an error. Please see the log file '$LogFile'"
        Exit 1
    }
    1 { 
        # One additional drive will be found when capturing an operating system WITHOUT a 'System Reserved' boot partition (e.g. Windows 7, Windows Server 2008 R2)
        [string]$DriveLetterToDefrag = "$($Drives):"
        DS_WriteLog "S" "One new drive was added: $DriveLetterToDefrag" $LogFile
    }
    2 {
        # Two additional drives will be found when capturing an operating system WITH a 'System Reserved' boot partition (e.g. Windows 10, Windows Server 2016)
        DS_WriteLog "I" "Two new drives were added. Checking which drive requires offline defragmentation" $LogFile
        foreach ( $Drive in $Drives ) {
            DS_WriteLog "I" "Checking drive $($Drive):" $LogFile
            if ( (Get-WmiObject -Class win32_volume -Filter "DriveLetter = '$($Drive):'").Label -like "*Reserved*") {
                DS_WriteLog "I" "     Drive $($Drive): is the 'System Reserved' drive. This one does not require offline defragmentation" $LogFile
            } else {
                DS_WriteLog "I" "     Drive $($Drive): is the primary partition and requires offline defragmentation" $LogFile
                [string]$DriveLetterToDefrag = "$($Drive):"
            }
        }
    }
    default { 
        DS_WriteLog "E" "More than two new drives were added. This script is not able to verify which drive needs offline defragmentation" $LogFile
        DS_WriteLog "E" "This script will now quit." $LogFile
        Write-Host "The script ended in an error. Please see the log file '$LogFile'"
        Exit 1
    }
}

DS_WriteLog "-" "" $LogFile

# Retrieve available disks after VHD(X) mount
$DisksAvailableAfterVHDXMount = (Get-WmiObject Win32_DiskDrive).Index | sort-object
DS_WriteLog "I" "Retrieve available disks (after mount): $([string]$DisksAvailableAfterVHDXMount)" $LogFile
DS_WriteLog "-" "" $LogFile

# Check which disks were added after mounting the VHD(X) file
try {
    [array]$Disks = (Compare-Object $DisksAvailableBeforeVHDXMount $DisksAvailableAfterVHDXMount).InputObject
} catch {
    DS_WriteLog "E" "No additional disks were detected. Please check if the VHD(X) file was properly mounted" $LogFile
    Write-Host "The script ended in an error. Please see the log file '$LogFile'"
    Exit 1
}

# Continue check which disks were added after mounting the VHD(X) file and determine the one that needs to have its unique ID reset
$DiskCount = $Disks.Count
switch ($DiskCount)
{        
    0 { 
        DS_WriteLog "E" "No new disks were added. Apparently the VHD(X) mount did not succeed" $LogFile
        DS_WriteLog "E" "This script will now quit." $LogFile
        Write-Host "The script ended in an error. Please see the log file '$LogFile'"
        Exit 1
    }
    1 { 
        # One additional disk was found
        [string]$NewDisk = "$($Disks)"
        DS_WriteLog "I" "One new disk was added. Disk number: $NewDisk" $LogFile
    }
    default {
        DS_WriteLog "E" "More than one new disks were added. This script is not able to verify which disk needs to have its unique ID reset" $LogFile
        DS_WriteLog "E" "This script will now quit." $LogFile
        Write-Host "The script ended in an error. Please see the log file '$LogFile'"
        Exit 1
    }
}

DS_WriteLog "-" "" $LogFile

# Defrag VHD(X)
Write-Host "Defragging the VHD(X) mounted to drive $DriveLetterToDefrag..."
DS_WriteLog "I" "Defrag the VHD(X) mounted to drive $DriveLetterToDefrag" $LogFile
DS_WriteLog "I" "Execute command: Start-Process -WindowStyle hidden -FilePath ""$DefragFile"" -ArgumentList $DriveLetterToDefrag -Wait -PassThru" $LogFile
$Process = Start-Process -WindowStyle hidden -FilePath $DefragFile -ArgumentList $DriveLetterToDefrag -Wait -PassThru
$ProcessExitCode = $Process.ExitCode
if ($ProcessExitCode -eq 0 ) {
    DS_WriteLog "I" "Defrag drive $DriveLetterToDefrag completed successfully" $LogFile
    Write-Host "Defragging drive $DriveLetterToDefrag completed successfully"
} else {
    DS_WriteLog "E" "An error occurred while attempting to defrag drive $DriveLetterToDefrag (error: $ProcessExitCode)" $LogFile
    Write-Host "The script ended in an error. Please see the log file '$LogFile'"
    Exit 1
}

DS_WriteLog "-" "" $LogFile

# Create second temporary diskpart configuration file (to be able to un-mount the VHD(X))
$DiskpartFile = "$env:TEMP\diskpart_unmount.txt"
$Line1 = "SELECT DISK=""$($NewDisk)"""
$Line2 = "UNIQUEID DISK ID=""$($UniqueDiskIDHex)"""
$Line3 = "SELECT VDISK FILE=""$($VHDFileToDefrag)"""
$Line4 = "DETACH VDISK"
DS_WriteLog "I" "Create temporary diskpart file to unmount the VHD(X) file ($DiskpartFile)" $LogFile
try {
    New-Item $DiskpartFile -type file -force | Out-Null
    DS_WriteLog "I" "Temporary diskpart file was created successfully" $LogFile
    Add-Content $DiskpartFile $Line1
    DS_WriteLog "I" "The line $Line1 was added successfully" $LogFile
    Add-Content $DiskpartFile $Line2
    DS_WriteLog "I" "The line $Line2 was added successfully" $LogFile
    Add-Content $DiskpartFile $Line3
    DS_WriteLog "I" "The line $Line3 was added successfully" $LogFile
    Add-Content $DiskpartFile $Line4
    DS_WriteLog "I" "The line $Line4 was added successfully" $LogFile
} Catch {
    DS_WriteLog "E" "An error occurred trying to create the diskpart configuration file (error: $($Error[0]))" $LogFile
    Write-Host "The script ended in an error. Please see the log file '$LogFile'"
    Exit 1
}

DS_WriteLog "-" "" $LogFile

# (Re)set unique disk ID and un-mount VHD(X) (using diskpart)
DS_WriteLog "I" "(Re)set unique disk ID and un-mount (detach) the VHD(X) (using diskpart.exe)" $LogFile
$Process = Start-Process -WindowStyle hidden -FilePath 'diskpart.exe' -ArgumentList "/S $DiskpartFile" -Wait -PassThru
$ProcessExitCode = $Process.ExitCode
if ($ProcessExitCode -eq 0 ) {
    DS_WriteLog "I" "The diskpart command executed successfully" $LogFile
    Write-Host "The VHD(X) file was unmounted successfully"
} else {
    DS_WriteLog "E" "An error occurred executing the diskpart command (error: $ProcessExitCode)" $LogFile
    Write-Host "The script ended in an error. Please see the log file '$LogFile'"
    Exit 1
}

# Enable File Security  
Remove-Item env:\SEE_MASK_NOZONECHECKS

DS_WriteLog "-" "" $LogFile
DS_WriteLog "I" "End of script" $LogFile
Write-Host "End of script"
